#pragma once

#include <set>
#include <memory>
#include <algorithm>
#include <iostream>

#include "tools/mathutils.h"
#include "core/note.h"

template <class TNote, class = std::enable_if_t<std::is_base_of<Note, TNote>::value>>
class Timeline
{
public:
    explicit Timeline() :
        _visibility_offset(0),
        _current_offset(0)
    {}

    typedef typename std::set<TNote*>::const_iterator Iterator;

    void recalculate(const microsec& offset)
    {
        _current_offset = offset;
        expire(_first_visible_note);
        expire(_last_visible_note);
        expire(_top_note);

        if (!_timeline.empty())
        {
            Iterator head_iterator = _timeline.begin();

            while (!isExpired(head_iterator))
            {
                if ((*head_iterator)->offset() >= offset)
                {
                    Iterator pre_head = head_iterator;
                    --pre_head;

                    _top_note = !isExpired(pre_head) && (*pre_head)->isActive(offset)
                              ? pre_head
                              : head_iterator;
                    break;
                }

                ++head_iterator;
            }

            if (isExpired(_top_note))
                _top_note = _timeline.begin();
        }

        fetchVisibleNotes();
    }

    void setNotes(const std::set<TNote*, NotePtrCompt>& notes, const microsec& visibility)
    {
        _visibility_offset = visibility;
        _timeline = std::move(notes);

        recalculate(_current_offset);

        if (isExpired(_top_note))
            return;

        fetchVisibleNotes();
    }

    void insertNote(TNote* note)
    {
        _top_note = _timeline.insert(note).first;
        recalculate(_current_offset);
        update(_current_offset);
    }

    void insertNotes(const std::set<TNote*, NotePtrCompt>& notes)
    {
        _timeline.insert(notes.begin(), notes.end());
        recalculate(_current_offset);
        update(_current_offset);

    }

    inline void clear()
    {
        for (auto& note : _timeline)
            delete note;

        _timeline.clear();
    }

    void update(const microsec& offset)
    {
        _current_offset = offset;
        updateTopNote(_current_offset);
        updateVisibleSprites(_current_offset);
    }

    std::pair<Iterator, Iterator> getVisibleNotes() const
    {
        if (nothingToDraw())
            return std::pair<Iterator, Iterator>();

        return std::make_pair(_first_visible_note, _last_visible_note);
    }

    void fetchVisibleNotes()
    {
        findLastVisibleNote(_current_offset);
        findFirstVisibleNote();
    }

    void findLastVisibleNote(const microsec& music_offset)
    {
        Iterator note_iterator = _top_note;
        while (!isExpired(note_iterator) && isVisiblyClose(note_iterator, music_offset))
        {
            if (nothingToDraw())
                _first_visible_note = note_iterator;

            auto note = *note_iterator;

            if (!note->isInGame())
                note->putToGame(music_offset);

            ++note_iterator;
        }

        _last_visible_note = note_iterator;
    }

    void findFirstVisibleNote()
    {
        if (nothingToDraw())
            return;

        auto note_iterator = _first_visible_note;
        while (note_iterator != _last_visible_note)
        {
            auto note = *note_iterator;
            if (note->shouldRemove())
            {
                ++_first_visible_note;
            }

            ++note_iterator;
        }
    }

    Iterator getActiveNote(const microsec& music_offset) noexcept
    {
        Iterator return_note = _timeline.end();

        auto note_iterator = _top_note;
        while (!isExpired(note_iterator))
        {
            const auto& note = *note_iterator;
            if (note->isActive(music_offset))
            {
                return_note = note_iterator;
                break;
            }
            else if (note->offset() > music_offset)
                break;

            ++note_iterator;
        }

        return return_note;
    }

    inline Iterator getNoteBy(const microsec& music_offset) noexcept
    {
        return std::find_if(_timeline.begin(), _timeline.end(),
                            [music_offset](const auto& note)
                            {
                                return note->offset() == music_offset;
                            });
    }

    inline bool isExpired(const Iterator& iterator) const noexcept
    {
        return iterator == _timeline.end();
    }

    inline void expire(Iterator& iterator) noexcept
    {
        iterator = _timeline.end();
    }

private:
    std::set<TNote*, NotePtrCompt> _timeline;
    microsec _visibility_offset;
    microsec _current_offset;

    inline void updateTopNote(const microsec& music_offset) noexcept
    {
        if ((*_top_note)->offset() < music_offset     //
         && _top_note == _first_visible_note           // Maybe simplify
         && !(*_top_note)->isActive(music_offset))   //
        {
            ++_top_note;
        }
    }

    void updateVisibleSprites(const microsec& music_offset)
    {
        if (nothingToDraw())
            return;

        std::for_each(_first_visible_note, _last_visible_note,
                      [&music_offset](const auto& note)
                      {
                          note->update(music_offset);
                      });
    }

    inline bool isVisiblyClose(const Iterator& iterator, const microsec& music_offset) const noexcept
    {
        return ((*iterator)->offset() - _visibility_offset) <= music_offset;
    }

    inline bool nothingToDraw() const noexcept
    {
        return isExpired(_first_visible_note);
    }

    Iterator _top_note;
    Iterator _last_visible_note;
    Iterator _first_visible_note;
};
